/**
 * Copyright (c) 2004-2013, Abel Hegedus, Zoltan Ujhelyi and Daniel Varro
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors:
 *   Abel Hegedus, Zoltan Ujhelyi - initial API and implementation
 *   Peter Lunk - revised Transformation API structure for adapter support
 */
package org.eclipse.viatra.transformation.runtime.emf.transformation.batch;

import com.google.common.base.Predicate;
import com.google.common.collect.Iterators;
import com.google.common.collect.Sets;
import com.google.common.collect.UnmodifiableIterator;
import java.util.List;
import java.util.function.Consumer;
import org.eclipse.viatra.query.runtime.api.GenericQueryGroup;
import org.eclipse.viatra.query.runtime.api.IPatternMatch;
import org.eclipse.viatra.query.runtime.api.IQuerySpecification;
import org.eclipse.viatra.query.runtime.api.ViatraQueryEngine;
import org.eclipse.viatra.query.runtime.api.ViatraQueryMatcher;
import org.eclipse.viatra.transformation.evm.api.Activation;
import org.eclipse.viatra.transformation.evm.api.ConflictSetIterator;
import org.eclipse.viatra.transformation.evm.api.IExecutor;
import org.eclipse.viatra.transformation.evm.api.RuleEngine;
import org.eclipse.viatra.transformation.evm.api.RuleSpecification;
import org.eclipse.viatra.transformation.evm.api.event.EventFilter;
import org.eclipse.viatra.transformation.evm.api.resolver.ScopedConflictSet;
import org.eclipse.viatra.transformation.runtime.emf.filters.MatchParameterFilter;
import org.eclipse.viatra.transformation.runtime.emf.rules.BatchTransformationRuleGroup;
import org.eclipse.viatra.transformation.runtime.emf.rules.batch.BatchTransformationRule;
import org.eclipse.viatra.transformation.runtime.emf.transformation.batch.BatchTransformation;
import org.eclipse.xtext.xbase.lib.CollectionLiterals;
import org.eclipse.xtext.xbase.lib.Conversions;
import org.eclipse.xtext.xbase.lib.Functions.Function1;
import org.eclipse.xtext.xbase.lib.IterableExtensions;
import org.eclipse.xtext.xbase.lib.Pair;

/**
 * Utility class for simple rule usage
 * 
 * @author Abel Hegedus, Zoltan Ujhelyi, Peter Lunk
 */
@SuppressWarnings("all")
public class BatchTransformationStatements {
  private final ViatraQueryEngine queryEngine;
  
  private final RuleEngine ruleEngine;
  
  private final IExecutor executor;
  
  BatchTransformationStatements(final BatchTransformation transformation, final IExecutor executor) {
    this.ruleEngine = transformation.ruleEngine;
    this.executor = executor;
    this.queryEngine = transformation.queryEngine;
  }
  
  /**
   * Executes the selected rule with the selected filter as long as there
   * are possible matches of its preconditions and the break condition is
   * not fulfilled. The matches are executed one-by-one, in case of conflicts
   * only one of the conflicting matches will cause an execution.
   */
  public <Match extends IPatternMatch> void fireUntil(final BatchTransformationRule<Match, ?> rule, final Predicate<Match> breakCondition) {
    String _name = rule.getName();
    String _plus = ("Fire_until_transaction_condition_ruleName: " + _name);
    this.executor.startExecution(_plus);
    final EventFilter<Match> filter = rule.getRuleSpecification().createEmptyFilter();
    this.<Match>fireUntil(rule.getRuleSpecification(), breakCondition, filter);
    String _name_1 = rule.getName();
    String _plus_1 = ("Fire_until_transaction_condition_ruleName: " + _name_1);
    this.executor.endExecution(_plus_1);
  }
  
  /**
   * Executes the selected rule with the selected filter as long as there
   * are possible matches of its precondition and the break condition is
   * not fulfilled. The matches are executed one-by-one, in case of conflicts
   * only one of the conflicting matches will cause an execution.
   */
  public <Match extends IPatternMatch> void fireUntil(final BatchTransformationRule<Match, ?> rule, final Predicate<Match> breakCondition, final Pair<String, ?>... filterParameters) {
    String _name = rule.getName();
    String _plus = ("Fire_until_transaction_filter_condition_ruleName: " + _name);
    this.executor.startExecution(_plus);
    RuleSpecification<Match> _ruleSpecification = rule.getRuleSpecification();
    MatchParameterFilter _matchParameterFilter = new MatchParameterFilter(filterParameters);
    this.<Match>fireUntil(_ruleSpecification, breakCondition, _matchParameterFilter);
    String _name_1 = rule.getName();
    String _plus_1 = ("Fire_until_transaction_filter_condition_ruleName: " + _name_1);
    this.executor.endExecution(_plus_1);
  }
  
  /**
   * Executes the selected rule with the selected filter as long as there
   * are possible matches of its precondition and the break condition is
   * not fulfilled. The matches are executed one-by-one, in case of conflicts
   * only one of the conflicting matches will cause an execution.
   */
  public <Match extends IPatternMatch> void fireUntil(final BatchTransformationRule<Match, ?> rule, final Predicate<Match> breakCondition, final EventFilter<? super Match> filter) {
    String _name = rule.getName();
    String _plus = ("Fire_until_transaction_filter_condition_ruleName: " + _name);
    this.executor.startExecution(_plus);
    this.<Match>fireUntil(rule.getRuleSpecification(), breakCondition, filter);
    String _name_1 = rule.getName();
    String _plus_1 = ("Fire_until_transaction_filter_condition_ruleName: " + _name_1);
    this.executor.endExecution(_plus_1);
  }
  
  /**
   * Executes the selected rules with the selected filter as long as there
   * are possible matches of any of their preconditions and the break condition is
   * not fulfilled. The matches are executed one-by-one, in case of conflicts
   * only one of the conflicting matches will cause an execution.
   */
  public void fireUntil(final BatchTransformationRuleGroup rules, final Predicate<IPatternMatch> breakCondition) {
    this.executor.startExecution("Fire_until_transaction_condition_ruleGroup");
    this.registerRules(rules);
    final ScopedConflictSet conflictSet = this.ruleEngine.createScopedConflictSet(rules.getFilteredRuleMap());
    this.<IPatternMatch>fireUntil(conflictSet, breakCondition);
    conflictSet.dispose();
    this.executor.endExecution("Fire_until_transaction_condition_ruleGroup");
  }
  
  /**
   * Executes the selected rule with the selected filter as long as there
   * are possible matches of its precondition. The matches are executed
   * one-by-one, in case of conflicts only one of the conflicting matches
   * will cause an execution.
   */
  public <Match extends IPatternMatch> void fireWhilePossible(final BatchTransformationRule<Match, ?> rule) {
    String _name = rule.getName();
    String _plus = ("Fire_while_possible_transaction_ruleName: " + _name);
    this.executor.startExecution(_plus);
    final Predicate<Match> _function = (Match it) -> {
      return false;
    };
    this.<Match>fireUntil(rule, _function);
    String _name_1 = rule.getName();
    String _plus_1 = ("Fire_while_possible_transaction_ruleName: " + _name_1);
    this.executor.endExecution(_plus_1);
  }
  
  /**
   * Executes the selected rule with the selected filter as long as there
   * are possible matches of its preconditions. The matches are executed
   * one-by-one, in case of conflicts only one of the conflicting matches
   * will cause an execution.
   */
  public <Match extends IPatternMatch> void fireWhilePossible(final BatchTransformationRule<Match, ?> rule, final EventFilter<? super Match> filter) {
    String _name = rule.getName();
    String _plus = ("Fire_while_possible_transaction_filter_ruleName: " + _name);
    this.executor.startExecution(_plus);
    final Predicate<Match> _function = (Match it) -> {
      return false;
    };
    this.<Match>fireUntil(rule, _function, filter);
    String _name_1 = rule.getName();
    String _plus_1 = ("Fire_while_possible_transaction_filter_ruleName: " + _name_1);
    this.executor.endExecution(_plus_1);
  }
  
  /**
   * Executes the selected rules with the selected filter as long as there
   * are possible matches of any of their preconditions. The matches are
   * executed one-by-one, in case of conflicts only one of the conflicting
   * matches will cause an execution.
   */
  public void fireWhilePossible(final BatchTransformationRuleGroup rules) {
    this.executor.startExecution("Fire_while_possible_transaction_ruleGroup");
    final Predicate<IPatternMatch> _function = (IPatternMatch it) -> {
      return false;
    };
    this.fireUntil(rules, _function);
    this.executor.endExecution("Fire_while_possible_transaction_ruleGroup");
  }
  
  /**
   * Executes the selected rule with the selected filter on its current match set of the precondition.
   */
  public <Match extends IPatternMatch> void fireAllCurrent(final BatchTransformationRule<Match, ?> rule) {
    String _name = rule.getName();
    String _plus = ("Fire_all_current_transaction_ruleName: " + _name);
    this.executor.startExecution(_plus);
    final EventFilter<Match> filter = rule.getRuleSpecification().createEmptyFilter();
    this.<Match>fireAllCurrent(rule.getRuleSpecification(), filter);
    String _name_1 = rule.getName();
    String _plus_1 = ("Fire_all_current_transaction_ruleName: " + _name_1);
    this.executor.endExecution(_plus_1);
  }
  
  /**
   * Executes the selected rule with the selected filter on its current match set of the precondition.
   */
  public <Match extends IPatternMatch> void fireAllCurrent(final BatchTransformationRule<Match, ?> rule, final Pair<String, ?>... parameterFilter) {
    String _name = rule.getName();
    String _plus = ("Fire_all_current_transaction_filter_ruleName: " + _name);
    this.executor.startExecution(_plus);
    RuleSpecification<Match> _ruleSpecification = rule.getRuleSpecification();
    MatchParameterFilter _matchParameterFilter = new MatchParameterFilter(parameterFilter);
    this.<Match>fireAllCurrent(_ruleSpecification, _matchParameterFilter);
    String _name_1 = rule.getName();
    String _plus_1 = ("Fire_all_current_transaction_filter_ruleName: " + _name_1);
    this.executor.endExecution(_plus_1);
  }
  
  /**
   * Executes the selected rule with the selected filter on its current match set of the precondition.
   */
  public <Match extends IPatternMatch> void fireAllCurrent(final BatchTransformationRule<Match, ?> rule, final EventFilter<? super Match> filter) {
    String _name = rule.getName();
    String _plus = ("Fire_all_current_transaction_filter_ruleName: " + _name);
    this.executor.startExecution(_plus);
    this.<Match>fireAllCurrent(rule.getRuleSpecification(), filter);
    String _name_1 = rule.getName();
    String _plus_1 = ("Fire_all_current_transaction_filter_ruleName: " + _name_1);
    this.executor.endExecution(_plus_1);
  }
  
  public <Match extends IPatternMatch> boolean registerRule(final RuleSpecification<Match> ruleSpecification) {
    return this.<Match>registerRule(ruleSpecification, ruleSpecification.createEmptyFilter());
  }
  
  public <Match extends IPatternMatch> boolean registerRule(final RuleSpecification<Match> ruleSpecification, final EventFilter<? super Match> filter) {
    return this.ruleEngine.<Match>addRule(ruleSpecification, filter);
  }
  
  public void registerRules(final BatchTransformationRuleGroup rules) {
    final Function1<BatchTransformationRule<?, ?>, IQuerySpecification<? extends ViatraQueryMatcher<? extends IPatternMatch>>> _function = (BatchTransformationRule<?, ?> it) -> {
      return it.getPrecondition();
    };
    final Iterable<IQuerySpecification<? extends ViatraQueryMatcher<? extends IPatternMatch>>> notNullPreconditions = IterableExtensions.<BatchTransformationRule<?, ?>, IQuerySpecification<? extends ViatraQueryMatcher<? extends IPatternMatch>>>map(IterableExtensions.<BatchTransformationRule<?, ?>>filterNull(rules), _function);
    GenericQueryGroup.of(Sets.<IQuerySpecification<?>>newHashSet(notNullPreconditions)).prepare(this.queryEngine);
    final Consumer<BatchTransformationRule<?, ?>> _function_1 = (BatchTransformationRule<?, ?> it) -> {
      EventFilter<?> _filter = it.getFilter();
      this.ruleEngine.addRule(it.getRuleSpecification(), ((EventFilter<IPatternMatch>) _filter));
    };
    IterableExtensions.<BatchTransformationRule<?, ?>>filterNull(rules).forEach(_function_1);
  }
  
  /**
   * Returns the number of current activations of the rule.
   * 
   * @since 1.5
   */
  public <Match extends IPatternMatch> int countAllCurrent(final BatchTransformationRule<Match, ?> rule) {
    int _xblockexpression = (int) 0;
    {
      final EventFilter<Match> filter = rule.getRuleSpecification().createEmptyFilter();
      _xblockexpression = this.<Match>countAllCurrent(rule.getRuleSpecification(), filter);
    }
    return _xblockexpression;
  }
  
  /**
   * Returns the number of current activations of the rule.
   * 
   * @since 1.5
   */
  public <Match extends IPatternMatch> int countAllCurrent(final BatchTransformationRule<Match, ?> rule, final Pair<String, ?>... parameterFilter) {
    RuleSpecification<Match> _ruleSpecification = rule.getRuleSpecification();
    MatchParameterFilter _matchParameterFilter = new MatchParameterFilter(parameterFilter);
    return this.<Match>countAllCurrent(_ruleSpecification, _matchParameterFilter);
  }
  
  /**
   * Returns the number of current activations of the rule.
   * 
   * @since 1.5
   */
  public <Match extends IPatternMatch> int countAllCurrent(final BatchTransformationRule<Match, ?> rule, final EventFilter<? super Match> filter) {
    return this.<Match>countAllCurrent(rule.getRuleSpecification(), filter);
  }
  
  public <Match extends IPatternMatch> boolean disposeRule(final RuleSpecification<Match> ruleSpecification) {
    return this.<Match>disposeRule(ruleSpecification, ruleSpecification.createEmptyFilter());
  }
  
  public <Match extends IPatternMatch> boolean disposeRule(final RuleSpecification<Match> ruleSpecification, final EventFilter<? super Match> filter) {
    return this.ruleEngine.<Match>removeRule(ruleSpecification, filter);
  }
  
  public void disposeRules(final BatchTransformationRuleGroup rules) {
    final Consumer<BatchTransformationRule<?, ?>> _function = (BatchTransformationRule<?, ?> it) -> {
      EventFilter<?> _filter = it.getFilter();
      this.ruleEngine.removeRule(it.getRuleSpecification(), ((EventFilter<IPatternMatch>) _filter));
    };
    IterableExtensions.<BatchTransformationRule<?, ?>>filterNull(rules).forEach(_function);
  }
  
  /**
   * Selects and fires an activation of the selected rule with no corresponding filter.</p>
   * 
   * <p><strong>Warning</strong>: the selection criteria is not specified - it is neither random nor controllable
   */
  public <Match extends IPatternMatch> void fireOne(final BatchTransformationRule<Match, ?> rule) {
    String _name = rule.getName();
    String _plus = ("Fire_one_transaction_ruleName: " + _name);
    this.executor.startExecution(_plus);
    final EventFilter<Match> filter = rule.getRuleSpecification().createEmptyFilter();
    this.<Match>fireOne(rule.getRuleSpecification(), filter);
    String _name_1 = rule.getName();
    String _plus_1 = ("Fire_one_transaction_ruleName: " + _name_1);
    this.executor.endExecution(_plus_1);
  }
  
  /**
   * Selects and fires an activation of the selected rule with a corresponding filter</p>
   * 
   * <p><strong>Warning</strong>: the selection criteria is not specified - it is neither random nor controllable
   */
  public <Match extends IPatternMatch> void fireOne(final BatchTransformationRule<Match, ?> rule, final Pair<String, ?>... parameterFilter) {
    String _name = rule.getName();
    String _plus = ("Fire_one_transaction_filter_ruleName: " + _name);
    this.executor.startExecution(_plus);
    RuleSpecification<Match> _ruleSpecification = rule.getRuleSpecification();
    MatchParameterFilter _matchParameterFilter = new MatchParameterFilter(parameterFilter);
    this.<Match>fireOne(_ruleSpecification, _matchParameterFilter);
    String _name_1 = rule.getName();
    String _plus_1 = ("Fire_one_transaction_filter_ruleName: " + _name_1);
    this.executor.endExecution(_plus_1);
  }
  
  /**
   * Selects and fires an activation of the selected rule with a corresponding filter.</p>
   * 
   * <p><strong>Warning</strong>: the selection criteria is not specified - it is neither random nor controllable
   */
  public <Match extends IPatternMatch> void fireOne(final BatchTransformationRule<Match, ?> rule, final EventFilter<? super Match> filter) {
    String _name = rule.getName();
    String _plus = ("Fire_one_transaction_filter_ruleName: " + _name);
    this.executor.startExecution(_plus);
    this.<Match>fireOne(rule.getRuleSpecification(), filter);
    String _name_1 = rule.getName();
    String _plus_1 = ("Fire_one_transaction_filter_ruleName: " + _name_1);
    this.executor.endExecution(_plus_1);
  }
  
  private <Match extends IPatternMatch> boolean fireOne(final RuleSpecification<Match> ruleSpecification, final EventFilter<? super Match> filter) {
    boolean _xblockexpression = false;
    {
      this.<Match>registerRule(ruleSpecification, filter);
      final ScopedConflictSet conflictSet = this.ruleEngine.<Match>createScopedConflictSet(ruleSpecification, filter);
      Activation<?> _head = IterableExtensions.<Activation<?>>head(conflictSet.getConflictingActivations());
      final UnmodifiableIterator<Activation<?>> iterator = Iterators.<Activation<?>>singletonIterator(((Activation<?>) _head));
      this.executor.execute(iterator);
      conflictSet.dispose();
      _xblockexpression = this.<Match>disposeRule(ruleSpecification, filter);
    }
    return _xblockexpression;
  }
  
  private <Match extends IPatternMatch> boolean fireAllCurrent(final RuleSpecification<Match> ruleSpecification, final EventFilter<? super Match> filter) {
    boolean _xblockexpression = false;
    {
      this.<Match>registerRule(ruleSpecification, filter);
      final ScopedConflictSet conflictSet = this.ruleEngine.<Match>createScopedConflictSet(ruleSpecification, filter);
      List<Activation<?>> list = CollectionLiterals.<Activation<?>>newArrayList(((Activation<?>[])Conversions.unwrapArray(conflictSet.getConflictingActivations(), Activation.class)));
      this.executor.execute(list.iterator());
      conflictSet.dispose();
      _xblockexpression = this.<Match>disposeRule(ruleSpecification, filter);
    }
    return _xblockexpression;
  }
  
  private <Match extends IPatternMatch> int countAllCurrent(final RuleSpecification<Match> ruleSpecification, final EventFilter<? super Match> filter) {
    this.<Match>registerRule(ruleSpecification, filter);
    final ScopedConflictSet conflictSet = this.ruleEngine.<Match>createScopedConflictSet(ruleSpecification, filter);
    int count = conflictSet.getConflictingActivations().size();
    conflictSet.dispose();
    this.<Match>disposeRule(ruleSpecification, filter);
    return count;
  }
  
  private <Match extends IPatternMatch> boolean fireUntil(final RuleSpecification<Match> ruleSpecification, final Predicate<Match> breakCondition, final EventFilter<? super Match> filter) {
    boolean _xblockexpression = false;
    {
      this.<Match>registerRule(ruleSpecification, filter);
      final ScopedConflictSet conflictSet = this.ruleEngine.<Match>createScopedConflictSet(ruleSpecification, filter);
      ScopedConflictSet _createScopedConflictSet = this.ruleEngine.<Match>createScopedConflictSet(ruleSpecification, filter);
      ConflictSetIterator _conflictSetIterator = new ConflictSetIterator(_createScopedConflictSet, breakCondition);
      this.executor.execute(_conflictSetIterator);
      conflictSet.dispose();
      _xblockexpression = this.<Match>disposeRule(ruleSpecification, filter);
    }
    return _xblockexpression;
  }
  
  private <Match extends IPatternMatch> void fireUntil(final ScopedConflictSet conflictSet, final Predicate<Match> breakCondition) {
    ConflictSetIterator _conflictSetIterator = new ConflictSetIterator(conflictSet, breakCondition);
    this.executor.execute(_conflictSetIterator);
  }
}
